"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const graphql_1 = require("graphql");
const fs_extra_1 = require("fs-extra");
const chokidar_1 = require("chokidar");
const glob = require("glob");
const ast_1 = require("graphql-tool-utilities/ast");
const print_1 = require("./print");
class Builder extends events_1.EventEmitter {
    constructor({ graphQLFiles, schemaPath }) {
        super();
        this.documentCache = new Map();
        this.globs = graphQLFiles;
        this.schemaPath = schemaPath;
    }
    on(event, handler) {
        return super.on(event, handler);
    }
    emit(event, ...args) {
        return super.emit(event, ...args);
    }
    run({ watch: watchGlobs = false } = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            this.watching = watchGlobs;
            const { globs } = this;
            const update = (file) => __awaiter(this, void 0, void 0, function* () {
                try {
                    yield this.updateDocumentForFile(file);
                }
                catch (error) {
                    this.emit('error', error);
                    return;
                }
                yield this.generate();
            });
            if (watchGlobs) {
                const documentWatcher = chokidar_1.watch(globs);
                documentWatcher.on('ready', () => {
                    documentWatcher.on('add', update);
                    documentWatcher.on('change', update);
                    documentWatcher.on('unlink', (file) => __awaiter(this, void 0, void 0, function* () {
                        this.removeDocumentForFile(file);
                        yield this.generate();
                    }));
                });
                const schemaWatcher = chokidar_1.watch(this.schemaPath);
                schemaWatcher.on('ready', () => {
                    schemaWatcher.on('change', () => __awaiter(this, void 0, void 0, function* () {
                        try {
                            yield this.updateSchema();
                            yield this.generate();
                        }
                        catch (error) {
                            return;
                        }
                    }));
                });
            }
            try {
                yield this.updateSchema();
            }
            catch (error) {
                this.emit('error', error);
                return;
            }
            try {
                yield Promise.all(glob
                    .sync(globs)
                    .map(this.updateDocumentForFile.bind(this)));
            }
            catch (error) {
                this.emit('error', error);
                return;
            }
            yield this.generate();
        });
    }
    generate() {
        return __awaiter(this, void 0, void 0, function* () {
            this.emit('start');
            let ast;
            try {
                ast = ast_1.compile(this.schema, graphql_1.concatAST(Array.from(this.documentCache.values())));
            }
            catch (error) {
                this.emit(error);
                return;
            }
            const fileMap = groupOperationsAndFragmentsByFile(ast);
            const buildResults = yield Promise.all(Object
                .keys(fileMap)
                .map((key) => __awaiter(this, void 0, void 0, function* () {
                const file = fileMap[key];
                const definition = print_1.printFile(file, ast);
                const definitionPath = `${file.path}.d.ts`;
                yield fs_extra_1.writeFile(definitionPath, definition);
                return {
                    documentPath: file.path,
                    definitionPath,
                    operations: file.operations,
                    fragments: file.fragments,
                };
            })));
            for (const buildResult of buildResults) {
                this.emit('build', buildResult);
            }
            this.emit('end');
        });
    }
    updateSchema() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const schemaJSON = yield fs_extra_1.readJSON(this.schemaPath, 'utf8');
                this.schema = graphql_1.buildClientSchema(schemaJSON.data);
            }
            catch (error) {
                const parseError = new Error(`Error parsing '${this.schemaPath}':\n\n${error.message.replace(/Syntax Error GraphQL \(.*?\) /, '')}`);
                throw parseError;
            }
        });
    }
    updateDocumentForFile(file) {
        return __awaiter(this, void 0, void 0, function* () {
            const contents = yield fs_extra_1.readFile(file, 'utf8');
            if (contents.trim().length === 0) {
                return;
            }
            const document = graphql_1.parse(new graphql_1.Source(contents, file));
            this.documentCache.set(file, document);
        });
    }
    removeDocumentForFile(file) {
        this.documentCache.delete(file);
    }
}
exports.Builder = Builder;
function groupOperationsAndFragmentsByFile({ operations, fragments }) {
    const map = {};
    Object
        .keys(operations)
        .forEach((name) => {
        const operation = operations[name];
        const file = map[operation.filePath] || { path: operation.filePath, operations: [], fragments: [] };
        file.operations.push(operation);
        map[operation.filePath] = file;
    });
    Object
        .keys(fragments)
        .forEach((name) => {
        const fragment = fragments[name];
        const file = map[fragment.filePath] || { path: fragment.filePath, operations: [], fragments: [] };
        file.fragments.push(fragment);
        map[fragment.filePath] = file;
    });
    return map;
}
